/***************************************************************\
*	chosc.c 						*
*	Non-linear oscilator					*
*	3 July 1996 John ffitch					*
*	Revised Nov 2003					*
\***************************************************************/

#include "csdl.h"
#include "chosc.h"

#ifdef never
typedef struct	{
	OPDS	h;
        MYFLT	*ar, *X0, *C;	/* The parameters */
        MYFLT   prev;
} SCHOSC;

typedef struct	{
	OPDS	h;
        MYFLT	*ar, *X0, *X1, *a, *b, *d, *C, *L;	/* The parameters */
        AUXCH	delay;		/* Buffer for old values */
	int	point;		/* Pointer to old values */
} CHOSC;
#endif

/* Y{n} = Y^2{n-1} - C */

void schoscset(SCHOSC *p)
{
    p->prev = *p->X0;
}


void schosc(SCHOSC *p)
{
    MYFLT	*ar;
    int		nsmps;
    MYFLT	C = *p->C;
    MYFLT	prev = p->prev;

    ar   = p->ar;
    nsmps = ksmps;		/* Number of points to calculate */
    if (C<FL(0.0)) C = FL(0.0);	/* C must be in range [0,2] */
    else if (C>FL(2.0)) C = FL(2.0);

    do {
      prev = prev*prev - C;
      *ar++ = prev;		/* Write output */
    } while (--nsmps);
    p->prev = prev;
} /* end schosc(p) */


/* X_n = a X_{n-1} + b X_{n-2} + d X_{n-L}^2 - C */

#define MAX_DELAY	(1024)

void choscset(CHOSC *p)
{
    int i;
    MYFLT *pp;

    if (p->delay.auxp == NULL) { /* get newspace    */
      auxalloc(MAX_DELAY*sizeof(MYFLT),&p->delay);
    }
    pp = (MYFLT*)p->delay.auxp;
    for (i=2; i<MAX_DELAY; i++) pp[i] = FL(0.0);/* Clear delays */
    pp[2] = *p->X1;
    pp[1] = *p->X0;
    p->point = 0;
} /* end choscset(p) */


void chosc(CHOSC *p)
{
    MYFLT	*ar;
    int		nsmps;
    int		point = p->point;
    int		nm1 = point-2;
    int		nm2 = point-2;
    int		nmL;
    MYFLT	ynm1, ynm2, ynmL;
    MYFLT	a = *p->a, b = *p->b, d = *p->d, C = *p->C;
    int		L = (int)(*p->L+0.5);
    MYFLT	*pp = (MYFLT*)p->delay.auxp;

    ar   = p->ar;
    nsmps = ksmps;		/* Number of points to calculate */
				/* L is k-rate so need to check */
    if (L < 1) L = 1;
    else if (L >= MAX_DELAY-2) {
      err_printf("Delay parameter too large at %d\n", L);
      L = MAX_DELAY-3;
    }
    nmL = point - L;
    if (nm1<0) nm1 += MAX_DELAY; /* Deal with the wrapping */
    if (nm2<0) nm2 += MAX_DELAY;
    if (nmL<0) nmL += MAX_DELAY;
    ynm1 = pp[nm1];	/* Pick up running values */
    ynm2 = pp[nm2];
    ynmL = pp[nmL];

    do {
      MYFLT yn;
      yn = a*ynm1 + b*ynm2 + d*ynmL*ynmL - C;
      *ar++ = yn;		/* Write output */
      if (++point >=MAX_DELAY) point -= MAX_DELAY;
      pp[point] = yn; /* and delay line */
      nm2 = nm1;
      nm1 = point;
      if (++nmL>=MAX_DELAY) nmL -= MAX_DELAY;
      ynm2 = ynm1;		/* Shuffle along */
      ynm1 = yn;
      ynmL = pp[nmL];
    } while (--nsmps);
    p->point = point;
} /* end chosc(p) */

/* #ifdef never */
#define S       sizeof

static OENTRY localops[] = {
  { "chosc1",  S(SCHOSC),  5,  "a",   "ik",	 (SUBR)schoscset, (SUBR)NULL, (SUBR)schosc},
  { "chosc2",  S(CHOSC),   5,  "a",   "iikkkkk", (SUBR)choscset,  (SUBR)NULL, (SUBR)chosc }
};

long opcode_size(void)
{
    return sizeof(localops);
}

OENTRY *opcode_init(GLOBALS *xx)
{
    pcglob = xx;
    return localops;
}
/* #endif */
