/* Spiral Synthesis
 * May 1999 - Bradley Bell <bradleyb@u.washington.edu>
 * this is a csound implementation of code from Tracy Lind Petersen's
 * "Spiral Synthesis" which can be read at:
 * http://staff.washington.edu/bradleyb/spiralsynth/
 *
 * don't forget to add something like these to entry.c:
 * 75a76
 * > #include "spiral.h"
 * 318a320
 * > void    spiralset(void*), spiral(void*);
 * 1058a1061
 * > { "spiral",S(SPIRAL),    5,    "aa",    "kkkk",  spiralset, NULL, spiral }
 *
 */

#include "csdl.h"
#include "spiral.h"
#include <math.h>

void spiralset(SPIRAL *p)
{
  p->phz = 0.0;
  p->dr = p->di = 0.0;
  p->xr = 1.0;
  p->mg2 = p->angl2 = FL(0.0);
}

void spiral(SPIRAL *p)
{
  int nsmps=ksmps;
  float *real = p->real;
  float *imag = p->imag;
  float *acar = p->acar;
  float *fcar = p->fcar;
  float *amod = p->amod;
  float *fmod = p->fmod;

  do {
    p->angl   = 2 * PI * *fcar / esr;	/* removed the ++ after fcar -- akozar, 112304 */
    p->phzdel = 2 * PI * *fmod / esr;	/* removed the ++ after fmod -- akozar, 112304 */

    p->mod   = *amod * sin(p->phz);		/* removed the ++ after amod -- akozar, 112304 */
    p->phz   += p->phzdel;
    p->mg2   = *acar + p->mod;		/* removed the ++ after acar -- akozar, 112304 */
    p->angl2 = p->angl + p->mod;

    p->a = p->mg2 * cos(p->angl2);
    p->b = p->mg2 * sin(p->angl2);

    p->yr = p->xr + p->a * p->dr - p->b * p->di;
    p->yi =         p->b * p->dr + p->a * p->di;

    p->dr = p->yr;
    p->di = p->yi;

    p->xr = 0;

    *real++ = p->yr;
    *imag++ = p->yi;
  } while (--nsmps);

}

/* added code for dynamic linkage -- akozar, 112304 */
#define S       sizeof

static OENTRY localops[] = {
  { "spiral", S(SPIRAL),    5,    "aa",    "kkkk",  (SUBR)spiralset, NULL, (SUBR)spiral }
};

long opcode_size(void)
{
    return sizeof(localops);
}

OENTRY *opcode_init(GLOBALS *xx)
{
    pcglob = xx;
    return localops;
}

